import json
import os
import logging
from collections import defaultdict
from datetime import datetime, timedelta, timezone
from typing import Dict, Any, List

from dotenv import load_dotenv
load_dotenv()
LOG_LEVEL = os.environ.get('LOG_LEVEL', 'INFO').upper()
HISTORY_FILE = os.environ.get('HISTORY_FILE', "playlist_history.json")
CACHE_EXPIRY_DAYS = int(os.environ.get('CACHE_EXPIRY_DAYS', 2))

import logging
logging.basicConfig(
    format='%(asctime)s %(levelname)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=LOG_LEVEL
)

from helpers.PlaylistBuilder import PlaylistBuilder
from helpers.ScoreSaberAPI import ScoreSaberAPI

def load_history() -> Dict[str, Any]:
    """
    Load the playlist history from a JSON file.
    
    :return: A dictionary containing the history.
    """
    if os.path.exists(HISTORY_FILE):
        with open(HISTORY_FILE, 'r') as f:
            return json.load(f)
    return {}

def save_history(history: Dict[str, Any]) -> None:
    """
    Save the playlist history to a JSON file.
    
    :param history: The history dictionary to save.
    """
    with open(HISTORY_FILE, 'w') as f:
        json.dump(history, f, indent=2)

def prompt_for_player_id(default_id: str = '76561199407393962') -> str:
    """
    Prompt the user to enter a ScoreSaber player ID.
    Uses a default ID if the user presses Enter without input.
    
    :param default_id: The default player ID to use.
    :return: The player ID entered by the user or the default.
    """
    prompt = f"Enter player ID (press Enter for default '{default_id}'): "
    player_id = input(prompt).strip() or default_id
    return player_id

def format_time_ago(time_difference: timedelta) -> str:
    """
    Format a timedelta object into a human-readable string.
    
    :param time_difference: The time difference to format.
    :return: A string representing the time difference (e.g., '5d', '2w').
    """
    days = time_difference.days
    
    if days < 7:
        return f"{days} days(s)"
    elif days < 30:
        weeks = days // 7
        return f"{weeks} week(s)"
    elif days < 365:
        months = days // 30
        return f"{months} month(s)"
    else:
        years = days // 365
        return f"{years} year(s)"

def normalize_difficulty(difficulty_name):
    # ScoreSaber difficulty naming convention
    difficulty_map = {
        '_ExpertPlus_SoloStandard': 'expertplus',
        '_Expert_SoloStandard': 'expert',
        '_Hard_SoloStandard': 'hard',
        '_Normal_SoloStandard': 'normal',
        '_Easy_SoloStandard': 'easy'
    }
    
    return difficulty_map.get(difficulty_name, difficulty_name.lower().replace('_solostandard', ''))

def playlist_strategy_oldscores(
    api: ScoreSaberAPI, 
    song_count: int = 10
) -> List[Dict[str, Any]]:
    """Build and format a list of songs based on old scores, avoiding reusing the same song+difficulty."""

    player_id = prompt_for_player_id()
    history = load_history()
    history.setdefault('oldscores', {})

    scores_data = api.get_player_scores(player_id, use_cache=True)
    all_scores = scores_data.get('playerScores', [])
    if not all_scores:
        logging.warning(f"No scores found for player ID {player_id}.")
        return []
    logging.debug(f"Found {len(all_scores)} scores for player ID {player_id}.")

    # Sort scores by timeSet in ascending order (oldest first)
    all_scores.sort(key=lambda x: x['score'].get('timeSet', ''))

    difficulty_groups = defaultdict(list)
    no_stars_group = []
    current_time = datetime.now(timezone.utc)

    for score in all_scores:
        leaderboard = score.get('leaderboard', {})
        stars = leaderboard.get('stars', 0)
        song_id = leaderboard.get('songHash')
        difficulty_raw = leaderboard.get('difficulty', {}).get('difficultyRaw', '')
        
        if not song_id or not difficulty_raw:
            logging.debug(f"Skipping score due to missing song_id or difficulty_raw: {score}")
            continue  # Skip if essential data is missing

        # Calculate time ago
        time_set_str = score['score'].get('timeSet')
        if not time_set_str:
            logging.debug(f"Skipping score due to missing timeSet: {score}")
            continue  # Skip if time_set is missing
        try:
            time_set = datetime.fromisoformat(time_set_str.replace('Z', '+00:00'))
        except ValueError as e:
            logging.error(f"Invalid time format for score ID {score['score'].get('id')}: {e}")
            continue
        time_difference = current_time - time_set
        time_ago = format_time_ago(time_difference)

        # Normalize the difficulty name
        difficulty = normalize_difficulty(difficulty_raw)

        # Check history to avoid reusing song+difficulty
        if song_id in history['oldscores'] and difficulty in history['oldscores'][song_id]:
            logging.debug(f"Skipping song {song_id} with difficulty {difficulty} as it's in history.")
            continue  # Skip if already used

        # Format the song data as expected by PlaylistBuilder
        song_dict = {
            'hash': song_id,
            'songName': leaderboard.get('songName', 'Unknown'),
            'difficulties': [
                {
                    'name': difficulty,
                    'characteristic': leaderboard.get('difficulty', {}).get('gameMode', 'Standard')
                }
            ]
        }

        # Group songs based on stars
        if stars is None or stars == 0:
            no_stars_group.append(song_dict)
        elif 0 < stars <= 3:
            difficulty_groups[0].append(song_dict)
        elif 4 <= stars < 6:
            difficulty_groups[1].append(song_dict)
        elif 6 <= stars < 7:
            difficulty_groups[2].append(song_dict)
        elif stars >= 7:
            difficulty_groups[3].append(song_dict)

    playlist_data = []

    # Define the order and counts for each difficulty group
    for difficulty_level, count in [(0, song_count), (1, song_count), (2, song_count), (3, song_count)]:
        group = difficulty_groups.get(difficulty_level, [])
        if not group:
            logging.debug(f"No songs found for difficulty level {difficulty_level}.")
            continue
        # Sort by difficulty name to ensure consistent ordering
        sorted_group = sorted(group, key=lambda x: x['difficulties'][0]['name'])
        logging.debug(f"Sorted group for difficulty {difficulty_level}: {[song['songName'] for song in sorted_group]}")
        # Select unique songs up to the desired count
        unique_songs = {}
        for song in sorted_group:
            song_id = song['hash']
            if song_id not in unique_songs:
                unique_songs[song_id] = song
                logging.debug(f"Selected song for playlist: {song['songName']} ({song['difficulties'][0]['name']})")
            if len(unique_songs) >= count:
                break
        playlist_data.extend(unique_songs.values())
        logging.debug(f"Added {len(unique_songs)} songs from difficulty level {difficulty_level} to playlist.")

    # Add no-stars songs separately
    if no_stars_group:
        unique_no_stars_songs = {}
        sorted_no_stars = sorted(no_stars_group, key=lambda x: x['difficulties'][0]['name'])  # Adjust sorting as needed
        logging.debug(f"Sorted no_stars_group: {[song['songName'] for song in sorted_no_stars]}")
        for song in sorted_no_stars:
            song_id = song['hash']
            if song_id not in unique_no_stars_songs:
                unique_no_stars_songs[song_id] = song
                logging.debug(f"Selected no-star song for playlist: {song['songName']} ({song['difficulties'][0]['name']})")
            if len(unique_no_stars_songs) >= 5:  # Limit to 5 no-stars songs
                break
        playlist_data.extend(unique_no_stars_songs.values())
        logging.debug(f"Added {len(unique_no_stars_songs)} no-star songs to playlist.")

    # Log if no songs were added
    if not playlist_data:
        logging.info("No new songs found to add to the playlist based on history.")
    else:
        # Log details of each song added to the playlist
        for song in playlist_data:
            song_name = song['songName']
            mapper = "Unknown"  # Since mapper info was used earlier but not included in the formatted data
            # To include mapper, you may need to adjust the formatting in playlist_strategy_oldscores()
            # For now, it's omitted or you need to include it differently
            difficulty = song['difficulties'][0]['name']
            logging.info(f"Song added: {song_name} ({difficulty}), mapped by {mapper}. Last played {time_ago} ago.")
        logging.info(f"Total songs added to playlist: {len(playlist_data)}")

    # Update history to avoid reusing the same song+difficulty
    for song in playlist_data:
        song_id = song['hash']
        difficulty_name = song['difficulties'][0]['name']
        history['oldscores'].setdefault(song_id, []).append(difficulty_name)
    save_history(history)

    return playlist_data

def create_playlist(strategy='oldscores') -> None:
    """
    Generate a playlist of songs from a range of difficulties, all with scores previously set a long time ago.
    The range of difficulties ensures that the first few songs are good for warming up.
    Avoids reusing the same song+difficulty in a playlist based on history.
    """
    api = ScoreSaberAPI(cache_expiry_days=CACHE_EXPIRY_DAYS)

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    playlist_name = f"{strategy}-{timestamp}"

    if strategy == 'oldscores':
        playlist_data = playlist_strategy_oldscores(api, song_count=10)

    if not playlist_data:
        logging.info("No new scores found to add to the playlist.")
        return

    PlaylistBuilder().create_playlist(
        playlist_data,
        playlist_title=playlist_name,
        playlist_author="SaberList Tool"
    )
